# -*- coding: UTF-8 -*-

import startReader
from common import logger
logger.location = 'cs'

import fnmatch
import os
import sys
import time
from common import config
from artix.common import servicelock
from unload.service.reader import Reader
import atexit
import tempfile
import signal
import Queue
import traceback
import threading
import json
from bottle import route, run, response, request
from httplib import INTERNAL_SERVER_ERROR

log = logger.getUnloadLogger('main')
log.info('Starting unloadService...')

lockFile = os.path.normpath('%s/locks/lockArtixUnloadService' % tempfile.gettempdir())
srvLock = servicelock.ServiceLock(lockFile, log)
atexit.register(srvLock.unlock)
signal.signal(signal.SIGTERM, srvLock.unlock)
r = Reader()
unloadTasks = Queue.PriorityQueue(5000)

if not srvLock.lock(False):
    log.error("Can not run service. This service is running")
    os._exit(1)  # выход без выполнения atexit


def findFile(directory, pattern):
    if os.path.exists(directory):
        for file in os.listdir(directory):
            if fnmatch.fnmatch(file.lower(), pattern.lower()):
                return os.path.join(directory, file)
    return None


@route('/exchangers/sales/unload', method='POST')
def unloadRestService():
    response.content_type = 'application/json'
    result = None
    try:
        task = json.loads(request.body.read(), encoding='utf8')
        priority = task.pop('priority', 10)
        unloadTasks.put((priority, task), block=True, timeout=60)
        mess = u'Task %s added to queue' % task
        log.info(mess)
        result = {u'resultCode': 0, u'info': mess}
    except Queue.Full, e:
        mess = u'Queue is full. Can not add task %s to queue.' % task
        log.warning(mess)
        result = {u'resultCode': 10, u'info': mess}
    except Exception, e:
        log.warning(u'Ошибка REST сервиса. %s' % str(e))
        log.warning(traceback.format_exc())
        response.status = INTERNAL_SERVER_ERROR
        result = {u'resultCode': 100, u'info': str(e)}
    return json.dumps(result, ensure_ascii=False)


def startRestServiceForCS():
    restServiceHost = config.getString('UNLOAD', 'restHost', '0.0.0.0')
    restServicePort = config.getString('UNLOAD', 'restPort', 28082)
    try:
        log.debug('runing REST unload sales service on %s:%s...' % (restServiceHost, restServicePort))
        t = threading.Thread(target=run, kwargs={'host': restServiceHost, 'port': restServicePort})
        t.daemon = True
        t.start()
    except Exception as e:
        log.warning(u'Ошибка запуска REST сервиса: %s' % e.message)
        log.warning(traceback.format_exc())


timeUnloadCurrentSales = 0
timeFindRequest = 0
timeNodeReread = 0
timeoutNodeReread = 300  # таймаут перечитывания настроек


def unloadByTask(nodes, unloadTask, notProcessedShops, taskCountPass):
    # пытаемся выполнить задачу на выгрузку не больше 20 раз, затем пропускаем ее
    if taskCountPass > 20:
        log.warning('Can not processed task %s for shops %s' % (unloadTask, notProcessedShops))
        log.warning('Number of passes to processing task exceeded. Get next task...')
        del notProcessedShops[:]
    if len(notProcessedShops) == 0:
        taskCountPass = 0
        if unloadTask:
            unloadTasks.task_done()
        try:
            priority, unloadTask = unloadTasks.get_nowait()
            log.info('Work with task %s' % unloadTask)
            notProcessedShops = [str(unloadTask.get('shopId'))] if (unloadTask.get('shopId')) else [str(id) for id in unloadTask.get('shopIds')] if (unloadTask.get('shopIds')) else [str(n.id) for n in nodes]
        except Queue.Empty:
            unloadTask = None
    else:
        taskCountPass += 1
        log.info("Not processed task %s for shops %s. Let's try again." % (unloadTask, notProcessedShops))
    if unloadTask:
        cashIds = [str(id) for id in unloadTask.get('cashIds')] if (unloadTask.get('cashIds')) else None
        for node in nodes:
            if str(node.id) in notProcessedShops:
                result = True
                if cashIds:
                    for cash in node.cashes:
                        if str(cash.cashId) in cashIds:
                            result = startReader.unload(node, workDirectory, shifts=unloadTask.get('shifts'),
                                    dates=unloadTask.get('dates'), granularity=unloadTask.get('granularity', 'document'),
                                    closedShifts=unloadTask.get('closedShifts', 0), changedShifts=unloadTask.get('changedShifts'), cashCode=cash.code)
                            if not result:
                                break
                else:
                    result = startReader.unload(node, workDirectory, shifts=unloadTask.get('shifts'),
                            dates=unloadTask.get('dates'), granularity=unloadTask.get('granularity', 'document'),
                            closedShifts=unloadTask.get('closedShifts', 0), changedShifts=unloadTask.get('changedShifts'), cashCode=unloadTask.get('cashCode'))
                if result:
                        notProcessedShops.remove(str(node.id))
    return unloadTask, notProcessedShops, taskCountPass


def readPropertiesFromConfig():
    unloadCurrentSalesInterval = config.getInt('UNLOAD', 'unloadCurrentSalesInterval', 600)
    unloadSalesInterval = config.getInt('UNLOAD', 'unloadSalesInterval', 20)
    closedShiftsByFileFlag = config.getInt('UNLOAD', 'closedShiftsByFileFlag', 2)
    workDirectory = config.getString('main', 'workDirectory', '%s/artixExchange/workDir' % tempfile.gettempdir())

    log.debug('UNLOAD.unloadCurrentSalesInterval - %s' % unloadCurrentSalesInterval)
    log.debug('UNLOAD.unloadSalesInterval - %s' % unloadSalesInterval)
    log.debug('UNLOAD.closedShiftsByFileFlag - %s' % closedShiftsByFileFlag)
    log.debug('main.workDirectory - %s' % workDirectory)
    
    return unloadCurrentSalesInterval, unloadSalesInterval, closedShiftsByFileFlag, workDirectory


if config.read(os.path.normpath(sys.path[0] + "/../config/cs/exchangers.ini")):
    nodes = []
    notProcessedTaskShops = []
    unloadTask = None
    taskCountPass = 0
    try:
        nodeGetter = __import__('unload.node.%s.nodeGetter' % config.getString('UNLOAD', 'nodeGetter', 'cs22'), fromlist='q')
        nodes = nodeGetter.getNodes()
        timeNodeReread = time.time()
    except Exception as e:
        log.error(u'Ошибка получения нод. %s' % str(e))
        log.error(traceback.format_exc())
    unloadCurrentSalesInterval, unloadSalesInterval, closedShiftsByFileFlag, workDirectory = readPropertiesFromConfig()
    startRestServiceForCS()
    passCountSending = 0
    while True:
        if timeNodeReread == 0 or (time.time() - timeoutNodeReread) > timeNodeReread:
            timeNodeReread = time.time()
            try:
                newNodes = nodeGetter.getNodes()
                nodes = newNodes
            except:
                pass
            unloadCurrentSalesInterval, unloadSalesInterval, closedShiftsByFileFlag, workDirectory = readPropertiesFromConfig()

        if unloadCurrentSalesInterval != 0:
            if timeUnloadCurrentSales == 0 or (time.time() - unloadCurrentSalesInterval) > timeUnloadCurrentSales:
                log.info(u'Выгрузка текущих продаж...')
                timeUnloadCurrentSales = time.time()
                for node in nodes:
                    for cash in node.cashes:
                        for converter in cash.converters:
                            if hasattr(converter, 'onlineSalePath') and converter.onlineSalePath:
                                log.info('Onload current sales: shopId=%s, cashCode=%s, onlineSalePath=%s' % (node.id, cash.code, converter.onlineSalePath))
                                startReader.unload(node, workDirectory, converterName=converter.name, cashCode=cash.code, closedShifts=1)
        #### Выгрузка по таску из REST'а
        unloadTask, notProcessedTaskShops, taskCountPass = unloadByTask(nodes, unloadTask, notProcessedTaskShops, taskCountPass)
        ##################################################################
        if timeFindRequest == 0 or (time.time() - unloadSalesInterval) > timeFindRequest:
            timeFindRequest = time.time()
            
            for node in nodes:
                for cash in node.cashes:
                    for converter in cash.converters:
                        if converter.flag:
                            requestFile = findFile(os.path.dirname(converter.flag), os.path.basename(converter.flag))
                            if requestFile:
                                log.info(u'Выгрузка продаж по запросу файла-флага %s' % converter.flag)
                                startReader.unload(node, workDirectory, request=requestFile, converterName=converter.name, cashCode=cash.code, closedShifts=closedShiftsByFileFlag)
        if passCountSending < 5:
            passCountSending += 1
        else:
            passCountSending = 0
            for node in nodes:
                if findFile(workDirectory+'/%s/unload' % node.id, 'ready'):
                    r.send(node, workDirectory)
                if findFile(workDirectory+'/%s/unload' % node.id, 'readyOnline'):
                    r.send(node, workDirectory, closedShifts=1)
        time.sleep(1)
    # ждем пока все задачи на выгрузку обработаются
    unloadTasks.join()
else:
    log.error("Configuration file 'exchangers.ini' not found.")
